#include <string>
#include <fstream>
#include <map>

#include <iostream>
#include <fstream>
#include <sstream>


#include "basepairs.hh"
#include "math.h"

using namespace std;

/*

TODO

Error handling when reading input files!


*/



/* Details for the data structures in BasePairs

basepairs and their probabilities are stored in

  
- arc_probs_
  This is a (sparse) matrix, where we store basepair-probabilities.
  arc_probs_ contains entries for *all* probabilities given in the input

The follopwing structures are used by the alignment procedure
for traversing the basepairs. These structures contain only the relevant arcs
due to the basepair filtering.

- arcs_
  This is a (sparse) matrix, where we can lookup the arc-index
  for a pair of positions. This gives also information whether there is a relevant arc.
- arc_vec_
  The arcs indexed with the same indices as in arcs_
  for each arc, we record the left and right end, the weight, and the "stacking" weight 
- left_
  left_[i] is a list of arcs that have left end i, list only indices
- right_
  right_[j] is a list of arcs that have right end j, list only indices

there is a further structure for stacking probabilites

arc_probs_stack_
this is in analogy to arc_probs_stack_ and contains the probs P[(i,j) /\ (i+1,j-1)]


*/


/*

New strategy for reading basepairs

two stages:

* while reading insert into SparseMatrix
  
* afterwards construct traversal structures


*/





/*! compare right ends of arcs (for sorting) */
bool operator < (const LeftAdjEntry &e1,const LeftAdjEntry &e2) {
    return e1.right()<e2.right();
}

/*! compare left ends of arcs (for sorting) */
bool operator < (const RightAdjEntry &e1,const RightAdjEntry &e2) {
    return e1.left()>e2.left();
}

/*! resize data structures that allow fast acces to an arc arcs by
  its left end, its right end, or its left and right end
*/

void BasePairs::resize(uint s) {
    s_=s;
  
    left_.resize(s);
    right_.resize(s);
  

}

double BasePairs::prob_min() const {
    return params->min_prob;
}

double BasePairs::prob_exp() const {
    if (params->opt_exp_prob) {
	return  params->exp_prob;
    } else {
	assert(len_ != 0);
	return params->prob_exp_f(len_);
    }
}

int BasePairs::probToWeight(double p) const {
    double pe=prob_exp();
    
    return
	(int)round(params->structure_weight*
		   (1-log(p)/log(pe))
		   );
    
    /*
      Score soll 0 sein für p=pe und structure_weight für p=1
      
      ==> Normalisierung log(p/pe) / log(1/pe) 
      //                  [=1-log(p)/log(pe)]

    */
    
}


void BasePairs::sortAdjLists() {

    for (uint i=0; i<left_.size(); ++i)
	sort(left_[i].begin(),left_[i].end());
    for (uint i=0; i<right_.size(); ++i)
	sort(right_[i].begin(),right_[i].end());

}


string seqname_from_filename(const string &s) {
    int i;
    int j;
    for (i=s.length()-1; i>=0 && s[i]!='/'; i--)
	;

    for (j=i+1; j<s.length() && s[j]!='.'; j++)
	;

    string name=s.substr(i+1,j-i-1);

    if (name.length()>3 && name.substr(name.length()-3,3) == "_dp")
	name=name.substr(0,name.length()-3);
    
    return name;
}



//! generate the lists arc_vec_, left_, right_ from the matrices arc_probs_ and arc_probs_stack_
void BasePairs::generateBPLists() {
    
  // traverse the entries in the pro matrices

  // handle all arcs
  for (arc_prob_matrix_t::const_iterator it=arc_probs_.begin() ;
       arc_probs_.end()!=it ;
       ++it) {
    
    int i = it->first.first;
    int j = it->first.second;
    double p = it->second.first;    
    double cond_p;

    if (get_arc_prob(i+1,j-1)>0) { 
      cond_p=it->second.second / get_arc_prob(i+1,j-1);
    } else {
      cond_p=0;
    }
    
    
    if (p >= params->min_prob || cond_p >= params->min_prob) {
      // std::cout << i << " " << j << " " << p << " " << cond_p << std::endl;
    
      register_arc(i,j,p,cond_p);
    }
  }
  
  sortAdjLists();

}

//! registers an arc (i,j) with probability prob
//! and conditional probability cond_prob
//!
void BasePairs::register_arc(int i, int j, double prob, double cond_prob) {
  assert(i<j);
    
  score_t weight = probToWeight(prob);
  score_t cond_weight = probToWeight(cond_prob);
    
    int idx=arc_vec_.size();
    
    arc_vec_.push_back(Arc(idx,i,j,weight,cond_weight));
    
    left_[i].push_back(LeftAdjEntry(arc_vec_,idx));
    right_[j].push_back(RightAdjEntry(arc_vec_,idx));
        
    arcs_.set(i,j,idx);
}




void BasePairs::read(char *filename, Sequence &seq, bool stacking) {
  
  stacking_=stacking;

    ifstream in(filename);
    if (! in.good()) {
	cerr << "Cannot read "<<filename<<endl;
	exit(-1);
    }
    
    string s;
    // read first line and decide about file-format
    in >> s;
    in.close();
    if (s == "%!PS-Adobe-3.0") {
	// try reading as dot.ps file (as generated by RNAfold)
	readPS(filename, seq);
    } else {
	if (stacking_) {
	    std::cerr << "Not supported: PP input with stacking probs" << std::endl;
	    exit(-1);
	}

	// try reading as PP-file (proprietary format, which is easy to read and contains pair probs)
	readPP(filename, seq);
    }
}

void BasePairs::readPS(char *filename, Sequence &seq) {
    ifstream in(filename);

    string s;
    while (in >> s && s!="/sequence")
	;
    in >> s; in >> s;

    string seqstr="";
    while (in >> s && s!=")") {    
	s = s.substr(0,s.size()-1); // chop of last character
	// cout << s <<endl;
	seqstr+=s;
    }
    
    seq.append_row(seqname_from_filename(filename),seqstr);
    
    len_ = seq.length();
    
    resize(len_+1);
    
    char buf[1024];
    
    while (in.getline(buf,1024)) {
	
	if (strcmp(buf+strlen(buf)-4,"ubox")==0
	    || strcmp(buf+strlen(buf)-4,"lbox")==0
	    ) {
	    char boxtype=buf[strlen(buf)-4];
	    
	    istringstream ss(buf);
	    int i,j;
	    double p;
	    ss >> i >> j >> p;
	    
	    p*=p;
	    
	    assert(i<j);
	    	    
	    if (boxtype=='u') {
	      set_arc_prob(i,j,p);
	    }
	    else if (stacking_ && boxtype=='l') { // read a stacking probability
	      set_arc_stack_prob(i,j,p);
	    }
	}
    }

    generateBPLists();
}

void BasePairs::readPP(char *filename, Sequence &seq) {
    ifstream in(filename);
    
    string name;
    string seqstr;
    

    // ----------------------------------------
    // read sequence/alignment

    map<string,string> seq_map;
    
    const int BUF_SIZE=4096;
    char buf[BUF_SIZE];
    while (in.getline(buf,BUF_SIZE) && strcmp(buf,"#")!=0 ) {
	if (strlen(buf)>0 && buf[0]!=' ') {
	    istringstream in(buf);
	    in >> name >> seqstr;
	    if (name != "SCORE:") { // ignore the (usually first) line that begins with SCORE:
		seq_map[name]+=seqstr;
	    }
	}
    }
    
    for (map<string,string>::iterator it=seq_map.begin(); it!=seq_map.end(); ++it) {
	// std::cout << "SEQ: " << it->first << " " << it->second << std::endl;
	seq.append_row(it->first,it->second);
    }
    

    // ----------------------------------------
    // read base pairs
    
    int i,j;
    double p;
    double stack_p;
    double cond_p;

    len_=seq.length();

    // std::cout << "LEN: " << len_<<std::endl;
    
    resize(len_+1);

    while( in.getline(buf,BUF_SIZE) ) {
      istringstream in(buf);
      
      in >>i>>j>>p ;

      assert(i<j);
      
      if (stacking_) {
	in >> stack_p;
	
	if ( arc_probs_(i+1,j-1).first != 0 )
	  cond_p = stack_p / arc_probs_(i+1,j-1).first;
	else 
	  cond_p = 0;
      }

      
      arc_probs_.set(i,j,arc_prob_matrix_t::val_t(p,cond_p));
    }
    
    generateBPLists();
}
