#ifndef BASEPAIRS_HH
#define BASEPAIRS_HH


#include<iostream>

#include <vector>
#include <ext/hash_map>
#include <assert.h>

#include "base.hh"
#include "sequence.hh"

#include "params.hh"



/* in order to save some space we use a hash_map for storing the entries of the sparse
   matrices that allow fast acces to the basepairs (i,j) in a structure.
*/

namespace __gnu_cxx {
    //! defines a new hash function for pairs of ints
    template<>
    struct hash<std::pair<size_t,size_t> >
    {
	size_t
	operator()(std::pair<size_t,size_t> p) const
	{ return p.first<<(sizeof(size_t)/2) | p.second; }
    };
}

//! represents a sparse 2D matrix of entries val_t
//! implements the matrix by a hash map
template <class T>
class SparseMatrix {
public:
  typedef T val_t;
  typedef std::pair<size_t,size_t> key_t;
protected:
  typedef __gnu_cxx::hash_map<key_t,val_t > map_t;

    val_t def;
  
    map_t the_map;
public:

  typedef typename map_t::const_iterator const_iterator;


    SparseMatrix(val_t deflt) : the_map(),def(deflt) {}
  
    const val_t & operator() (size_t i, size_t j) const {
	const_iterator it = the_map.find(key_t(i,j));
	if ( it == the_map.end() ) 
	    return def;
	else 
	    return it->second;
    }
  
  
    void set(size_t i, size_t j, const val_t &val) {
	the_map[std::pair<int,int>(i,j)]=val;
    }
  
  
  const_iterator begin() const {
    return the_map.begin();
  }
  const_iterator end() const {
    return the_map.end();
  }
  
};


//! represents a base pair; stores base pair left end, right end, and weight
class Arc {
  int idx_;
    int left_;
    int right_;
    score_t w_;
    score_t w_stack_; //!< extra weight for stacked bp
public:
  
  //! constructor
  Arc(int idx, int left, int right, score_t w, score_t w_stack):
    idx_(idx), left_(left), right_(right), w_(w), w_stack_(w_stack)
  {}
  
  int left() const {return left_;}
  int right() const {return right_;}
  score_t weight() const {return w_;}
  score_t stack_weight() const {return w_stack_;}
  int idx() const {return idx_;}
  
  //void set_weight(score_t w) {w_=w;}
  //void set_stack_weight(score_t w_stack) {w_stack_=w_stack;}


};

typedef std::vector<Arc> arc_vec_t;


//! entry of an adjacency list
class AdjEntry {
protected:
  arc_vec_t av_;
  int idx_;
public:
  AdjEntry(arc_vec_t &av, int idx): av_(av),idx_(idx) {}
  
  score_t weight() const {return av_[idx_].weight();}
  score_t stack_weight() const {return av_[idx_].stack_weight();}
  int idx() const {return av_[idx_].idx();}
};

class LeftAdjEntry : public AdjEntry {
public:
  LeftAdjEntry(arc_vec_t &av, int idx): AdjEntry(av,idx) {}
  int right() const {return av_[idx_].right();}
};

class RightAdjEntry : public AdjEntry {
public:
  RightAdjEntry(arc_vec_t &av, int idx): AdjEntry(av,idx) {}
  int left() const {return av_[idx_].left();}
};


// ============================================================
// class BasePairs
/*! stores and maintains the list of base pairs together with their score
  contributions in arc matches.
*/
// ============================================================
class BasePairs
{
protected:
  BasePairsParams *params;
  
  bool stacking_;

public:
	
    /* types for data structures for the access of an arc in the structure,
       by its right end, its left end, or left and right end 
	
       the access structure is implemented as a hash map,
    */
	
    typedef std::vector<LeftAdjEntry> LeftAdjList;
    typedef std::vector<RightAdjEntry> RightAdjList;

    typedef SparseMatrix<int> arc_matrix_t;
    typedef SparseMatrix<std::pair<double, double> > arc_prob_matrix_t;


protected:
    std::vector<LeftAdjList> left_;
    std::vector<RightAdjList> right_;
    arc_vec_t arc_vec_;

    arc_matrix_t arcs_;
    arc_prob_matrix_t arc_probs_; //!< array for all arc probabilities and conditional probabilites
  //!< the array is used
  //!< when reading in the probabilities
  //!< and for merging probs during pp-output
  
    
    void resize(uint s);
    
  void generateBPLists();

    void sortAdjLists();
  

    int s_; //!< the sequence
    int len_; //!< the sequence length
public:
  BasePairs(BasePairsParams *bpp):
    params(bpp), s_(0), len_(0), 
    arcs_(-1), arc_probs_(arc_prob_matrix_t::val_t(0,0))
  {}

    //! registers a basepair (i,j) with prob and stack_prob
    //! maintains the basepair access data structures
    void register_arc(int i, int j, double p, double stack_p);    
  
    //! returns probability of basepair (i,j)
    double get_arc_prob(int i, int j) const {
	assert(i<j); 
	return arc_probs_(i,j).first;
    }

    //! returns stacking probability of basepairs (i,j) and (i+1,j-1)
    double get_arc_stack_prob(int i, int j) const {
	assert(i<j); 
	return arc_probs_(i,j).second;
    }


    //! sets probability of basepair (i,j)
    double set_arc_prob(int i, int j, double p) {
	assert(i<j); 
	arc_prob_matrix_t::val_t ps = arc_probs_(i,j);
	ps.first=p;
	arc_probs_.set(i,j,ps);
    }

    //! sets stacking probability of basepairs (i,j) and (i+1,j-1)
    double set_arc_stack_prob(int i, int j, double p) {
	assert(i<j); 
	arc_prob_matrix_t::val_t ps = arc_probs_(i,j);
	ps.second=p;
	arc_probs_.set(i,j,ps);
    }

    //! returns the list of arcs with right end i
    const LeftAdjList & left_adjlist(int i) const {return left_[i];}
	
    //! returns the list of arcs with left end i
    const RightAdjList & right_adjlist(int i) const {return right_[i];}

    //! accesses basepair by (i,j)
    const Arc &arc(int i,int j) const {return arc_vec_[arcs_(i,j)];}
    //! returns whether basepair (i,j) exitsts
    bool exists_arc(int i,int j) const {return -1 != arcs_(i,j);}

    //! returns number of basepairs in the object
    int num_bps() const {return arc_vec_.size();}
	
    //! returns weight that corresponds to probability p
    int probToWeight(double p) const;
	
    //! read baepairs and sequence from a pp-format file
    //! pp is a proprietary format of LocARNA
    //! which starts with the sequence and then simply
    //! lists the arcs (i,j) with their probabilities p
    void readPP(char *filename, Sequence &seq);
    
    //! read baepairs and sequence from a dp-format ps file
    //! dp is written by RNAfold -p
    void readPS(char *filename, Sequence &seq);
    
    //! read baepairs and sequence from a file
    //! (try to autodetect file format)
    //!@param filename the input file
    //!@param seq output parameter for the sequence
    //!@param stacking reads also stacking probabilities
    void read(char *filename, Sequence &seq, bool stacking);

    double prob_exp() const; //!< return expected probability
    double prob_min() const; //!< return minimal probability

    //! returns the base pair parameters that were used to init the object
    const BasePairsParams *get_params() const {return params;}
};

#endif
